# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: continual/rpc/logging/v1/logging.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from google.api import annotations_pb2 as google_dot_api_dot_annotations__pb2
from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from continual.rpc.management.v1 import management_types_pb2 as continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='continual/rpc/logging/v1/logging.proto',
  package='continual.rpc.logging.v1',
  syntax='proto3',
  serialized_options=b'ZCgithub.com/continualiq/continual/continual/rpc/logging/v1;loggingv1',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n&continual/rpc/logging/v1/logging.proto\x12\x18\x63ontinual.rpc.logging.v1\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x1cgoogle/api/annotations.proto\x1a\x1bgoogle/protobuf/empty.proto\x1a\x32\x63ontinual/rpc/management/v1/management_types.proto\"\'\n\x13LogHeartbeatRequest\x12\x10\n\x08resource\x18\x01 \x01(\t\"l\n\x15LogExperimentsRequest\x12\x15\n\rmodel_version\x18\x01 \x01(\t\x12<\n\x0b\x65xperiments\x18\x02 \x03(\x0b\x32\'.continual.rpc.management.v1.Experiment\"i\n\x11LogMetricsRequest\x12\x15\n\rmodel_version\x18\x01 \x01(\t\x12=\n\x07metrics\x18\x02 \x03(\x0b\x32,.continual.rpc.management.v1.TrainingMetrics\"\x9f\x01\n LogEvaluationModelMetricsRequest\x12\x15\n\rmodel_version\x18\x01 \x01(\t\x12\x64\n evaluation_model_version_metrics\x18\x17 \x01(\x0b\x32:.continual.rpc.management.v1.EvaluationModelVersionMetrics\"l\n\x16LogDatasetStatsRequest\x12\x10\n\x08resource\x18\x01 \x01(\t\x12@\n\rdataset_stats\x18\x02 \x01(\x0b\x32).continual.rpc.management.v1.DatasetStats\"z\n\x19LogPredictionStatsRequest\x12\x18\n\x10\x62\x61tch_prediction\x18\x01 \x01(\t\x12\x43\n\x10prediction_stats\x18\x02 \x01(\x0b\x32).continual.rpc.management.v1.DatasetStats\"\x88\x01\n GenerateArtifactUploadURLRequest\x12\x10\n\x08resource\x18\x01 \x01(\t\x12\x0c\n\x04path\x18\x02 \x01(\t\x12\x0c\n\x04type\x18\x03 \x01(\t\x12\x11\n\tmime_type\x18\x04 \x01(\t\x12\x10\n\x08metadata\x18\x05 \x01(\t\x12\x11\n\tresumable\x18\x06 \x01(\x08\"i\n!GenerateArtifactUploadURLResponse\x12\x0b\n\x03url\x18\x01 \x01(\t\x12\x37\n\x08\x61rtifact\x18\x02 \x01(\x0b\x32%.continual.rpc.management.v1.Artifact\"f\n\x14LogDataChecksRequest\x12\x10\n\x08resource\x18\x01 \x01(\t\x12<\n\x0b\x64\x61ta_checks\x18\x02 \x01(\x0b\x32\'.continual.rpc.management.v1.DataChecks\"p\n\x16LogModelVersionRequest\x12@\n\rmodel_version\x18\x01 \x01(\x0b\x32).continual.rpc.management.v1.ModelVersion\x12\x14\n\x0cupdate_paths\x18\x02 \x03(\t\"y\n\x19LogBatchPredictionRequest\x12\x46\n\x10\x62\x61tch_prediction\x18\x01 \x01(\x0b\x32,.continual.rpc.management.v1.BatchPrediction\x12\x14\n\x0cupdate_paths\x18\x02 \x03(\t\"\"\n\x12GetArtifactRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"9\n\x18\x42\x61tchGetArtifactsRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\r\n\x05names\x18\x02 \x03(\t\"U\n\x19\x42\x61tchGetArtifactsResponse\x12\x38\n\tartifacts\x18\x01 \x03(\x0b\x32%.continual.rpc.management.v1.Artifact\"%\n\x15\x44\x65leteArtifactRequest\x12\x0c\n\x04name\x18\x01 \x01(\t2\xa3\x0c\n\nLoggingAPI\x12l\n\x0cLogHeartbeat\x12-.continual.rpc.logging.v1.LogHeartbeatRequest\x1a\x16.google.protobuf.Empty\"\x15\x82\xd3\xe4\x93\x02\x0f\"\r/v1/heartbeat\x12r\n\x0eLogExperiments\x12/.continual.rpc.logging.v1.LogExperimentsRequest\x1a\x16.google.protobuf.Empty\"\x17\x82\xd3\xe4\x93\x02\x11\"\x0f/v1/experiments\x12\x66\n\nLogMetrics\x12+.continual.rpc.logging.v1.LogMetricsRequest\x1a\x16.google.protobuf.Empty\"\x13\x82\xd3\xe4\x93\x02\r\"\x0b/v1/metrics\x12\x8e\x01\n\x19LogEvaluationModelMetrics\x12:.continual.rpc.logging.v1.LogEvaluationModelMetricsRequest\x1a\x16.google.protobuf.Empty\"\x1d\x82\xd3\xe4\x93\x02\x17\"\x15/v1/evaluationMetrics\x12u\n\x0fLogDatasetStats\x12\x30.continual.rpc.logging.v1.LogDatasetStatsRequest\x1a\x16.google.protobuf.Empty\"\x18\x82\xd3\xe4\x93\x02\x12\"\x10/v1/datasetStats\x12~\n\x12LogPredictionStats\x12\x33.continual.rpc.logging.v1.LogPredictionStatsRequest\x1a\x16.google.protobuf.Empty\"\x1b\x82\xd3\xe4\x93\x02\x15\"\x13/v1/predictionStats\x12\x62\n\x0bGetArtifact\x12,.continual.rpc.logging.v1.GetArtifactRequest\x1a%.continual.rpc.management.v1.Artifact\x12|\n\x11\x42\x61tchGetArtifacts\x12\x32.continual.rpc.logging.v1.BatchGetArtifactsRequest\x1a\x33.continual.rpc.logging.v1.BatchGetArtifactsResponse\x12Y\n\x0e\x44\x65leteArtifact\x12/.continual.rpc.logging.v1.DeleteArtifactRequest\x1a\x16.google.protobuf.Empty\x12\x94\x01\n\x19GenerateArtifactUploadURL\x12:.continual.rpc.logging.v1.GenerateArtifactUploadURLRequest\x1a;.continual.rpc.logging.v1.GenerateArtifactUploadURLResponse\x12o\n\rLogDataChecks\x12..continual.rpc.logging.v1.LogDataChecksRequest\x1a\x16.google.protobuf.Empty\"\x16\x82\xd3\xe4\x93\x02\x10\"\x0e/v1/dataChecks\x12x\n\x0fLogModelVersion\x12\x30.continual.rpc.logging.v1.LogModelVersionRequest\x1a\x16.google.protobuf.Empty\"\x1b\x82\xd3\xe4\x93\x02\x15\"\x13/v1/logModelVersion\x12\x83\x01\n\x12LogBatchPrediction\x12\x33.continual.rpc.logging.v1.LogBatchPredictionRequest\x1a\x16.google.protobuf.Empty\" \x82\xd3\xe4\x93\x02\x1a\"\x18/v1/batchPredictionStateBEZCgithub.com/continualiq/continual/continual/rpc/logging/v1;loggingv1b\x06proto3'
  ,
  dependencies=[google_dot_protobuf_dot_timestamp__pb2.DESCRIPTOR,google_dot_api_dot_annotations__pb2.DESCRIPTOR,google_dot_protobuf_dot_empty__pb2.DESCRIPTOR,continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2.DESCRIPTOR,])




_LOGHEARTBEATREQUEST = _descriptor.Descriptor(
  name='LogHeartbeatRequest',
  full_name='continual.rpc.logging.v1.LogHeartbeatRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='resource', full_name='continual.rpc.logging.v1.LogHeartbeatRequest.resource', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=212,
  serialized_end=251,
)


_LOGEXPERIMENTSREQUEST = _descriptor.Descriptor(
  name='LogExperimentsRequest',
  full_name='continual.rpc.logging.v1.LogExperimentsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='model_version', full_name='continual.rpc.logging.v1.LogExperimentsRequest.model_version', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='experiments', full_name='continual.rpc.logging.v1.LogExperimentsRequest.experiments', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=253,
  serialized_end=361,
)


_LOGMETRICSREQUEST = _descriptor.Descriptor(
  name='LogMetricsRequest',
  full_name='continual.rpc.logging.v1.LogMetricsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='model_version', full_name='continual.rpc.logging.v1.LogMetricsRequest.model_version', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='metrics', full_name='continual.rpc.logging.v1.LogMetricsRequest.metrics', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=363,
  serialized_end=468,
)


_LOGEVALUATIONMODELMETRICSREQUEST = _descriptor.Descriptor(
  name='LogEvaluationModelMetricsRequest',
  full_name='continual.rpc.logging.v1.LogEvaluationModelMetricsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='model_version', full_name='continual.rpc.logging.v1.LogEvaluationModelMetricsRequest.model_version', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='evaluation_model_version_metrics', full_name='continual.rpc.logging.v1.LogEvaluationModelMetricsRequest.evaluation_model_version_metrics', index=1,
      number=23, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=471,
  serialized_end=630,
)


_LOGDATASETSTATSREQUEST = _descriptor.Descriptor(
  name='LogDatasetStatsRequest',
  full_name='continual.rpc.logging.v1.LogDatasetStatsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='resource', full_name='continual.rpc.logging.v1.LogDatasetStatsRequest.resource', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='dataset_stats', full_name='continual.rpc.logging.v1.LogDatasetStatsRequest.dataset_stats', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=632,
  serialized_end=740,
)


_LOGPREDICTIONSTATSREQUEST = _descriptor.Descriptor(
  name='LogPredictionStatsRequest',
  full_name='continual.rpc.logging.v1.LogPredictionStatsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='batch_prediction', full_name='continual.rpc.logging.v1.LogPredictionStatsRequest.batch_prediction', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='prediction_stats', full_name='continual.rpc.logging.v1.LogPredictionStatsRequest.prediction_stats', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=742,
  serialized_end=864,
)


_GENERATEARTIFACTUPLOADURLREQUEST = _descriptor.Descriptor(
  name='GenerateArtifactUploadURLRequest',
  full_name='continual.rpc.logging.v1.GenerateArtifactUploadURLRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='resource', full_name='continual.rpc.logging.v1.GenerateArtifactUploadURLRequest.resource', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='path', full_name='continual.rpc.logging.v1.GenerateArtifactUploadURLRequest.path', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='type', full_name='continual.rpc.logging.v1.GenerateArtifactUploadURLRequest.type', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='mime_type', full_name='continual.rpc.logging.v1.GenerateArtifactUploadURLRequest.mime_type', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='metadata', full_name='continual.rpc.logging.v1.GenerateArtifactUploadURLRequest.metadata', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='resumable', full_name='continual.rpc.logging.v1.GenerateArtifactUploadURLRequest.resumable', index=5,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=867,
  serialized_end=1003,
)


_GENERATEARTIFACTUPLOADURLRESPONSE = _descriptor.Descriptor(
  name='GenerateArtifactUploadURLResponse',
  full_name='continual.rpc.logging.v1.GenerateArtifactUploadURLResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='url', full_name='continual.rpc.logging.v1.GenerateArtifactUploadURLResponse.url', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='artifact', full_name='continual.rpc.logging.v1.GenerateArtifactUploadURLResponse.artifact', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1005,
  serialized_end=1110,
)


_LOGDATACHECKSREQUEST = _descriptor.Descriptor(
  name='LogDataChecksRequest',
  full_name='continual.rpc.logging.v1.LogDataChecksRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='resource', full_name='continual.rpc.logging.v1.LogDataChecksRequest.resource', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='data_checks', full_name='continual.rpc.logging.v1.LogDataChecksRequest.data_checks', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1112,
  serialized_end=1214,
)


_LOGMODELVERSIONREQUEST = _descriptor.Descriptor(
  name='LogModelVersionRequest',
  full_name='continual.rpc.logging.v1.LogModelVersionRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='model_version', full_name='continual.rpc.logging.v1.LogModelVersionRequest.model_version', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_paths', full_name='continual.rpc.logging.v1.LogModelVersionRequest.update_paths', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1216,
  serialized_end=1328,
)


_LOGBATCHPREDICTIONREQUEST = _descriptor.Descriptor(
  name='LogBatchPredictionRequest',
  full_name='continual.rpc.logging.v1.LogBatchPredictionRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='batch_prediction', full_name='continual.rpc.logging.v1.LogBatchPredictionRequest.batch_prediction', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_paths', full_name='continual.rpc.logging.v1.LogBatchPredictionRequest.update_paths', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1330,
  serialized_end=1451,
)


_GETARTIFACTREQUEST = _descriptor.Descriptor(
  name='GetArtifactRequest',
  full_name='continual.rpc.logging.v1.GetArtifactRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.logging.v1.GetArtifactRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1453,
  serialized_end=1487,
)


_BATCHGETARTIFACTSREQUEST = _descriptor.Descriptor(
  name='BatchGetArtifactsRequest',
  full_name='continual.rpc.logging.v1.BatchGetArtifactsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.logging.v1.BatchGetArtifactsRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='names', full_name='continual.rpc.logging.v1.BatchGetArtifactsRequest.names', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1489,
  serialized_end=1546,
)


_BATCHGETARTIFACTSRESPONSE = _descriptor.Descriptor(
  name='BatchGetArtifactsResponse',
  full_name='continual.rpc.logging.v1.BatchGetArtifactsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='artifacts', full_name='continual.rpc.logging.v1.BatchGetArtifactsResponse.artifacts', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1548,
  serialized_end=1633,
)


_DELETEARTIFACTREQUEST = _descriptor.Descriptor(
  name='DeleteArtifactRequest',
  full_name='continual.rpc.logging.v1.DeleteArtifactRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.logging.v1.DeleteArtifactRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1635,
  serialized_end=1672,
)

_LOGEXPERIMENTSREQUEST.fields_by_name['experiments'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._EXPERIMENT
_LOGMETRICSREQUEST.fields_by_name['metrics'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._TRAININGMETRICS
_LOGEVALUATIONMODELMETRICSREQUEST.fields_by_name['evaluation_model_version_metrics'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._EVALUATIONMODELVERSIONMETRICS
_LOGDATASETSTATSREQUEST.fields_by_name['dataset_stats'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._DATASETSTATS
_LOGPREDICTIONSTATSREQUEST.fields_by_name['prediction_stats'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._DATASETSTATS
_GENERATEARTIFACTUPLOADURLRESPONSE.fields_by_name['artifact'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ARTIFACT
_LOGDATACHECKSREQUEST.fields_by_name['data_checks'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._DATACHECKS
_LOGMODELVERSIONREQUEST.fields_by_name['model_version'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._MODELVERSION
_LOGBATCHPREDICTIONREQUEST.fields_by_name['batch_prediction'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._BATCHPREDICTION
_BATCHGETARTIFACTSRESPONSE.fields_by_name['artifacts'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ARTIFACT
DESCRIPTOR.message_types_by_name['LogHeartbeatRequest'] = _LOGHEARTBEATREQUEST
DESCRIPTOR.message_types_by_name['LogExperimentsRequest'] = _LOGEXPERIMENTSREQUEST
DESCRIPTOR.message_types_by_name['LogMetricsRequest'] = _LOGMETRICSREQUEST
DESCRIPTOR.message_types_by_name['LogEvaluationModelMetricsRequest'] = _LOGEVALUATIONMODELMETRICSREQUEST
DESCRIPTOR.message_types_by_name['LogDatasetStatsRequest'] = _LOGDATASETSTATSREQUEST
DESCRIPTOR.message_types_by_name['LogPredictionStatsRequest'] = _LOGPREDICTIONSTATSREQUEST
DESCRIPTOR.message_types_by_name['GenerateArtifactUploadURLRequest'] = _GENERATEARTIFACTUPLOADURLREQUEST
DESCRIPTOR.message_types_by_name['GenerateArtifactUploadURLResponse'] = _GENERATEARTIFACTUPLOADURLRESPONSE
DESCRIPTOR.message_types_by_name['LogDataChecksRequest'] = _LOGDATACHECKSREQUEST
DESCRIPTOR.message_types_by_name['LogModelVersionRequest'] = _LOGMODELVERSIONREQUEST
DESCRIPTOR.message_types_by_name['LogBatchPredictionRequest'] = _LOGBATCHPREDICTIONREQUEST
DESCRIPTOR.message_types_by_name['GetArtifactRequest'] = _GETARTIFACTREQUEST
DESCRIPTOR.message_types_by_name['BatchGetArtifactsRequest'] = _BATCHGETARTIFACTSREQUEST
DESCRIPTOR.message_types_by_name['BatchGetArtifactsResponse'] = _BATCHGETARTIFACTSRESPONSE
DESCRIPTOR.message_types_by_name['DeleteArtifactRequest'] = _DELETEARTIFACTREQUEST
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

LogHeartbeatRequest = _reflection.GeneratedProtocolMessageType('LogHeartbeatRequest', (_message.Message,), {
  'DESCRIPTOR' : _LOGHEARTBEATREQUEST,
  '__module__' : 'continual.rpc.logging.v1.logging_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.logging.v1.LogHeartbeatRequest)
  })
_sym_db.RegisterMessage(LogHeartbeatRequest)

LogExperimentsRequest = _reflection.GeneratedProtocolMessageType('LogExperimentsRequest', (_message.Message,), {
  'DESCRIPTOR' : _LOGEXPERIMENTSREQUEST,
  '__module__' : 'continual.rpc.logging.v1.logging_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.logging.v1.LogExperimentsRequest)
  })
_sym_db.RegisterMessage(LogExperimentsRequest)

LogMetricsRequest = _reflection.GeneratedProtocolMessageType('LogMetricsRequest', (_message.Message,), {
  'DESCRIPTOR' : _LOGMETRICSREQUEST,
  '__module__' : 'continual.rpc.logging.v1.logging_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.logging.v1.LogMetricsRequest)
  })
_sym_db.RegisterMessage(LogMetricsRequest)

LogEvaluationModelMetricsRequest = _reflection.GeneratedProtocolMessageType('LogEvaluationModelMetricsRequest', (_message.Message,), {
  'DESCRIPTOR' : _LOGEVALUATIONMODELMETRICSREQUEST,
  '__module__' : 'continual.rpc.logging.v1.logging_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.logging.v1.LogEvaluationModelMetricsRequest)
  })
_sym_db.RegisterMessage(LogEvaluationModelMetricsRequest)

LogDatasetStatsRequest = _reflection.GeneratedProtocolMessageType('LogDatasetStatsRequest', (_message.Message,), {
  'DESCRIPTOR' : _LOGDATASETSTATSREQUEST,
  '__module__' : 'continual.rpc.logging.v1.logging_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.logging.v1.LogDatasetStatsRequest)
  })
_sym_db.RegisterMessage(LogDatasetStatsRequest)

LogPredictionStatsRequest = _reflection.GeneratedProtocolMessageType('LogPredictionStatsRequest', (_message.Message,), {
  'DESCRIPTOR' : _LOGPREDICTIONSTATSREQUEST,
  '__module__' : 'continual.rpc.logging.v1.logging_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.logging.v1.LogPredictionStatsRequest)
  })
_sym_db.RegisterMessage(LogPredictionStatsRequest)

GenerateArtifactUploadURLRequest = _reflection.GeneratedProtocolMessageType('GenerateArtifactUploadURLRequest', (_message.Message,), {
  'DESCRIPTOR' : _GENERATEARTIFACTUPLOADURLREQUEST,
  '__module__' : 'continual.rpc.logging.v1.logging_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.logging.v1.GenerateArtifactUploadURLRequest)
  })
_sym_db.RegisterMessage(GenerateArtifactUploadURLRequest)

GenerateArtifactUploadURLResponse = _reflection.GeneratedProtocolMessageType('GenerateArtifactUploadURLResponse', (_message.Message,), {
  'DESCRIPTOR' : _GENERATEARTIFACTUPLOADURLRESPONSE,
  '__module__' : 'continual.rpc.logging.v1.logging_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.logging.v1.GenerateArtifactUploadURLResponse)
  })
_sym_db.RegisterMessage(GenerateArtifactUploadURLResponse)

LogDataChecksRequest = _reflection.GeneratedProtocolMessageType('LogDataChecksRequest', (_message.Message,), {
  'DESCRIPTOR' : _LOGDATACHECKSREQUEST,
  '__module__' : 'continual.rpc.logging.v1.logging_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.logging.v1.LogDataChecksRequest)
  })
_sym_db.RegisterMessage(LogDataChecksRequest)

LogModelVersionRequest = _reflection.GeneratedProtocolMessageType('LogModelVersionRequest', (_message.Message,), {
  'DESCRIPTOR' : _LOGMODELVERSIONREQUEST,
  '__module__' : 'continual.rpc.logging.v1.logging_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.logging.v1.LogModelVersionRequest)
  })
_sym_db.RegisterMessage(LogModelVersionRequest)

LogBatchPredictionRequest = _reflection.GeneratedProtocolMessageType('LogBatchPredictionRequest', (_message.Message,), {
  'DESCRIPTOR' : _LOGBATCHPREDICTIONREQUEST,
  '__module__' : 'continual.rpc.logging.v1.logging_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.logging.v1.LogBatchPredictionRequest)
  })
_sym_db.RegisterMessage(LogBatchPredictionRequest)

GetArtifactRequest = _reflection.GeneratedProtocolMessageType('GetArtifactRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETARTIFACTREQUEST,
  '__module__' : 'continual.rpc.logging.v1.logging_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.logging.v1.GetArtifactRequest)
  })
_sym_db.RegisterMessage(GetArtifactRequest)

BatchGetArtifactsRequest = _reflection.GeneratedProtocolMessageType('BatchGetArtifactsRequest', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETARTIFACTSREQUEST,
  '__module__' : 'continual.rpc.logging.v1.logging_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.logging.v1.BatchGetArtifactsRequest)
  })
_sym_db.RegisterMessage(BatchGetArtifactsRequest)

BatchGetArtifactsResponse = _reflection.GeneratedProtocolMessageType('BatchGetArtifactsResponse', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETARTIFACTSRESPONSE,
  '__module__' : 'continual.rpc.logging.v1.logging_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.logging.v1.BatchGetArtifactsResponse)
  })
_sym_db.RegisterMessage(BatchGetArtifactsResponse)

DeleteArtifactRequest = _reflection.GeneratedProtocolMessageType('DeleteArtifactRequest', (_message.Message,), {
  'DESCRIPTOR' : _DELETEARTIFACTREQUEST,
  '__module__' : 'continual.rpc.logging.v1.logging_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.logging.v1.DeleteArtifactRequest)
  })
_sym_db.RegisterMessage(DeleteArtifactRequest)


DESCRIPTOR._options = None

_LOGGINGAPI = _descriptor.ServiceDescriptor(
  name='LoggingAPI',
  full_name='continual.rpc.logging.v1.LoggingAPI',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=1675,
  serialized_end=3246,
  methods=[
  _descriptor.MethodDescriptor(
    name='LogHeartbeat',
    full_name='continual.rpc.logging.v1.LoggingAPI.LogHeartbeat',
    index=0,
    containing_service=None,
    input_type=_LOGHEARTBEATREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=b'\202\323\344\223\002\017\"\r/v1/heartbeat',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='LogExperiments',
    full_name='continual.rpc.logging.v1.LoggingAPI.LogExperiments',
    index=1,
    containing_service=None,
    input_type=_LOGEXPERIMENTSREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=b'\202\323\344\223\002\021\"\017/v1/experiments',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='LogMetrics',
    full_name='continual.rpc.logging.v1.LoggingAPI.LogMetrics',
    index=2,
    containing_service=None,
    input_type=_LOGMETRICSREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=b'\202\323\344\223\002\r\"\013/v1/metrics',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='LogEvaluationModelMetrics',
    full_name='continual.rpc.logging.v1.LoggingAPI.LogEvaluationModelMetrics',
    index=3,
    containing_service=None,
    input_type=_LOGEVALUATIONMODELMETRICSREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=b'\202\323\344\223\002\027\"\025/v1/evaluationMetrics',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='LogDatasetStats',
    full_name='continual.rpc.logging.v1.LoggingAPI.LogDatasetStats',
    index=4,
    containing_service=None,
    input_type=_LOGDATASETSTATSREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=b'\202\323\344\223\002\022\"\020/v1/datasetStats',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='LogPredictionStats',
    full_name='continual.rpc.logging.v1.LoggingAPI.LogPredictionStats',
    index=5,
    containing_service=None,
    input_type=_LOGPREDICTIONSTATSREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=b'\202\323\344\223\002\025\"\023/v1/predictionStats',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetArtifact',
    full_name='continual.rpc.logging.v1.LoggingAPI.GetArtifact',
    index=6,
    containing_service=None,
    input_type=_GETARTIFACTREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ARTIFACT,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='BatchGetArtifacts',
    full_name='continual.rpc.logging.v1.LoggingAPI.BatchGetArtifacts',
    index=7,
    containing_service=None,
    input_type=_BATCHGETARTIFACTSREQUEST,
    output_type=_BATCHGETARTIFACTSRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='DeleteArtifact',
    full_name='continual.rpc.logging.v1.LoggingAPI.DeleteArtifact',
    index=8,
    containing_service=None,
    input_type=_DELETEARTIFACTREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GenerateArtifactUploadURL',
    full_name='continual.rpc.logging.v1.LoggingAPI.GenerateArtifactUploadURL',
    index=9,
    containing_service=None,
    input_type=_GENERATEARTIFACTUPLOADURLREQUEST,
    output_type=_GENERATEARTIFACTUPLOADURLRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='LogDataChecks',
    full_name='continual.rpc.logging.v1.LoggingAPI.LogDataChecks',
    index=10,
    containing_service=None,
    input_type=_LOGDATACHECKSREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=b'\202\323\344\223\002\020\"\016/v1/dataChecks',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='LogModelVersion',
    full_name='continual.rpc.logging.v1.LoggingAPI.LogModelVersion',
    index=11,
    containing_service=None,
    input_type=_LOGMODELVERSIONREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=b'\202\323\344\223\002\025\"\023/v1/logModelVersion',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='LogBatchPrediction',
    full_name='continual.rpc.logging.v1.LoggingAPI.LogBatchPrediction',
    index=12,
    containing_service=None,
    input_type=_LOGBATCHPREDICTIONREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=b'\202\323\344\223\002\032\"\030/v1/batchPredictionState',
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_LOGGINGAPI)

DESCRIPTOR.services_by_name['LoggingAPI'] = _LOGGINGAPI

# @@protoc_insertion_point(module_scope)
