# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from continual.rpc.logging.v1 import logging_pb2 as continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2
from continual.rpc.management.v1 import management_types_pb2 as continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2
from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2


class LoggingAPIStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.LogHeartbeat = channel.unary_unary(
                '/continual.rpc.logging.v1.LoggingAPI/LogHeartbeat',
                request_serializer=continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.LogHeartbeatRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )
        self.LogExperiments = channel.unary_unary(
                '/continual.rpc.logging.v1.LoggingAPI/LogExperiments',
                request_serializer=continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.LogExperimentsRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )
        self.LogMetrics = channel.unary_unary(
                '/continual.rpc.logging.v1.LoggingAPI/LogMetrics',
                request_serializer=continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.LogMetricsRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )
        self.LogEvaluationModelMetrics = channel.unary_unary(
                '/continual.rpc.logging.v1.LoggingAPI/LogEvaluationModelMetrics',
                request_serializer=continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.LogEvaluationModelMetricsRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )
        self.LogDatasetStats = channel.unary_unary(
                '/continual.rpc.logging.v1.LoggingAPI/LogDatasetStats',
                request_serializer=continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.LogDatasetStatsRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )
        self.LogPredictionStats = channel.unary_unary(
                '/continual.rpc.logging.v1.LoggingAPI/LogPredictionStats',
                request_serializer=continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.LogPredictionStatsRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )
        self.GetArtifact = channel.unary_unary(
                '/continual.rpc.logging.v1.LoggingAPI/GetArtifact',
                request_serializer=continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.GetArtifactRequest.SerializeToString,
                response_deserializer=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2.Artifact.FromString,
                )
        self.BatchGetArtifacts = channel.unary_unary(
                '/continual.rpc.logging.v1.LoggingAPI/BatchGetArtifacts',
                request_serializer=continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.BatchGetArtifactsRequest.SerializeToString,
                response_deserializer=continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.BatchGetArtifactsResponse.FromString,
                )
        self.DeleteArtifact = channel.unary_unary(
                '/continual.rpc.logging.v1.LoggingAPI/DeleteArtifact',
                request_serializer=continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.DeleteArtifactRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )
        self.GenerateArtifactUploadURL = channel.unary_unary(
                '/continual.rpc.logging.v1.LoggingAPI/GenerateArtifactUploadURL',
                request_serializer=continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.GenerateArtifactUploadURLRequest.SerializeToString,
                response_deserializer=continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.GenerateArtifactUploadURLResponse.FromString,
                )
        self.LogDataChecks = channel.unary_unary(
                '/continual.rpc.logging.v1.LoggingAPI/LogDataChecks',
                request_serializer=continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.LogDataChecksRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )
        self.LogModelVersion = channel.unary_unary(
                '/continual.rpc.logging.v1.LoggingAPI/LogModelVersion',
                request_serializer=continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.LogModelVersionRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )
        self.LogBatchPrediction = channel.unary_unary(
                '/continual.rpc.logging.v1.LoggingAPI/LogBatchPrediction',
                request_serializer=continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.LogBatchPredictionRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )


class LoggingAPIServicer(object):
    """Missing associated documentation comment in .proto file."""

    def LogHeartbeat(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def LogExperiments(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def LogMetrics(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def LogEvaluationModelMetrics(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def LogDatasetStats(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def LogPredictionStats(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetArtifact(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def BatchGetArtifacts(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DeleteArtifact(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GenerateArtifactUploadURL(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def LogDataChecks(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def LogModelVersion(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def LogBatchPrediction(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_LoggingAPIServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'LogHeartbeat': grpc.unary_unary_rpc_method_handler(
                    servicer.LogHeartbeat,
                    request_deserializer=continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.LogHeartbeatRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'LogExperiments': grpc.unary_unary_rpc_method_handler(
                    servicer.LogExperiments,
                    request_deserializer=continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.LogExperimentsRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'LogMetrics': grpc.unary_unary_rpc_method_handler(
                    servicer.LogMetrics,
                    request_deserializer=continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.LogMetricsRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'LogEvaluationModelMetrics': grpc.unary_unary_rpc_method_handler(
                    servicer.LogEvaluationModelMetrics,
                    request_deserializer=continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.LogEvaluationModelMetricsRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'LogDatasetStats': grpc.unary_unary_rpc_method_handler(
                    servicer.LogDatasetStats,
                    request_deserializer=continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.LogDatasetStatsRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'LogPredictionStats': grpc.unary_unary_rpc_method_handler(
                    servicer.LogPredictionStats,
                    request_deserializer=continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.LogPredictionStatsRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'GetArtifact': grpc.unary_unary_rpc_method_handler(
                    servicer.GetArtifact,
                    request_deserializer=continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.GetArtifactRequest.FromString,
                    response_serializer=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2.Artifact.SerializeToString,
            ),
            'BatchGetArtifacts': grpc.unary_unary_rpc_method_handler(
                    servicer.BatchGetArtifacts,
                    request_deserializer=continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.BatchGetArtifactsRequest.FromString,
                    response_serializer=continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.BatchGetArtifactsResponse.SerializeToString,
            ),
            'DeleteArtifact': grpc.unary_unary_rpc_method_handler(
                    servicer.DeleteArtifact,
                    request_deserializer=continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.DeleteArtifactRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'GenerateArtifactUploadURL': grpc.unary_unary_rpc_method_handler(
                    servicer.GenerateArtifactUploadURL,
                    request_deserializer=continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.GenerateArtifactUploadURLRequest.FromString,
                    response_serializer=continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.GenerateArtifactUploadURLResponse.SerializeToString,
            ),
            'LogDataChecks': grpc.unary_unary_rpc_method_handler(
                    servicer.LogDataChecks,
                    request_deserializer=continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.LogDataChecksRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'LogModelVersion': grpc.unary_unary_rpc_method_handler(
                    servicer.LogModelVersion,
                    request_deserializer=continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.LogModelVersionRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'LogBatchPrediction': grpc.unary_unary_rpc_method_handler(
                    servicer.LogBatchPrediction,
                    request_deserializer=continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.LogBatchPredictionRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'continual.rpc.logging.v1.LoggingAPI', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class LoggingAPI(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def LogHeartbeat(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/continual.rpc.logging.v1.LoggingAPI/LogHeartbeat',
            continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.LogHeartbeatRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def LogExperiments(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/continual.rpc.logging.v1.LoggingAPI/LogExperiments',
            continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.LogExperimentsRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def LogMetrics(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/continual.rpc.logging.v1.LoggingAPI/LogMetrics',
            continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.LogMetricsRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def LogEvaluationModelMetrics(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/continual.rpc.logging.v1.LoggingAPI/LogEvaluationModelMetrics',
            continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.LogEvaluationModelMetricsRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def LogDatasetStats(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/continual.rpc.logging.v1.LoggingAPI/LogDatasetStats',
            continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.LogDatasetStatsRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def LogPredictionStats(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/continual.rpc.logging.v1.LoggingAPI/LogPredictionStats',
            continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.LogPredictionStatsRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetArtifact(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/continual.rpc.logging.v1.LoggingAPI/GetArtifact',
            continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.GetArtifactRequest.SerializeToString,
            continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2.Artifact.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def BatchGetArtifacts(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/continual.rpc.logging.v1.LoggingAPI/BatchGetArtifacts',
            continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.BatchGetArtifactsRequest.SerializeToString,
            continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.BatchGetArtifactsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def DeleteArtifact(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/continual.rpc.logging.v1.LoggingAPI/DeleteArtifact',
            continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.DeleteArtifactRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GenerateArtifactUploadURL(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/continual.rpc.logging.v1.LoggingAPI/GenerateArtifactUploadURL',
            continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.GenerateArtifactUploadURLRequest.SerializeToString,
            continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.GenerateArtifactUploadURLResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def LogDataChecks(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/continual.rpc.logging.v1.LoggingAPI/LogDataChecks',
            continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.LogDataChecksRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def LogModelVersion(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/continual.rpc.logging.v1.LoggingAPI/LogModelVersion',
            continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.LogModelVersionRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def LogBatchPrediction(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/continual.rpc.logging.v1.LoggingAPI/LogBatchPrediction',
            continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.LogBatchPredictionRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)
